﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace VolumeRayCaster
{
    interface World
    {
        bool GetPoint(Vector v);
    }

    class EmptyWorld : World
    {
        public bool GetPoint(Vector v)
        {
            return false;
        }
    }

    class Cube : World
    {
        public Cube(Vector a, Vector b)
            : base()
        {
            this.a = a;
            this.b = b;
        }
        public bool GetPoint(Vector v)
        {
            return v.x >= a.x && v.x <= b.x && v.y >= a.y && v.y <= b.y && v.z >= a.z && v.z <= b.z;
        }
        Vector a, b;
    }

    class Sphere : World
    {
        public Sphere(double radius)
        {
            this.radiusSqr = radius*radius;
        }
        public bool GetPoint(Vector v)
        {
            return v.LengthSqr <= radiusSqr;
        }
        double radiusSqr;
    }

    class Julia3D : World
    {
        public Julia3D(Vector c, int iters)
        {
            this.c = c;
            this.iters = iters;
        }
        Vector Power(Vector v)
        {
            double rxysqr = v.x * v.x + v.y * v.y;
            double a = 1 - v.z * v.z / rxysqr;
            return new Vector((v.x * v.x - v.y * v.y) * a, 2 * v.x * v.y * a, 2 * v.z * Math.Sqrt(rxysqr));
        }
        public bool GetPoint(Vector v)
        {
            for (int i = 0; i < iters; ++i)
            {
                v = Power(v);
                v += c;
                if (v.LengthSqr >= 4) return false;
            }
            return true;
        }
        Vector c;
        int iters;
    }

    class AddWorld : World
    {
        public AddWorld(World w1, World w2)
        {
            this.w1 = w1;
            this.w2 = w2;
        }
        public bool GetPoint(Vector v)
        {
            return w1.GetPoint(v) || w2.GetPoint(v);
        }
        World w1;
        World w2;
    }

    class SubtractWorld : World
    {
        public SubtractWorld(World w1, World w2)
        {
            this.w1 = w1;
            this.w2 = w2;
        }
        public bool GetPoint(Vector v)
        {
            return w1.GetPoint(v) && !w2.GetPoint(v);
        }
        World w1;
        World w2;
    }

    class IntersectWorld : World
    {
        public IntersectWorld(World w1, World w2)
        {
            this.w1 = w1;
            this.w2 = w2;
        }
        public bool GetPoint(Vector v)
        {
            return w1.GetPoint(v) && w2.GetPoint(v);
        }
        World w1;
        World w2;
    }

    class HalfspaceWorld : World
    {
        public HalfspaceWorld(double A, double B, double C, double D)
        {
            this.A = A;
            this.B = B;
            this.C = C;
            this.D = D;
        }
        public bool GetPoint(Vector v)
        {
            return v.x * A + v.y * B + v.z * C + D < 0;
        }
        double A,B,C,D;
    }
}
